#include "mamalleg.h"
#include "driver.h"
#include <math.h>
#include <pc.h>
#include "gen15khz.h"
#include "mode15k.h"




static void vesa_scroll(int x, int y);

/* from video.c */
extern int video_depth, video_attributes, video_orientation;
extern float video_fps;
extern int width_for_mode15k, height_for_mode15k;

extern int scanlines;
extern int stretch;
extern int center_x;
extern int center_y;
extern int video_sync;
extern int use_triplebuf;
extern int gfx_xoffset;
extern int gfx_yoffset;

//extern int gfx_display_lines;
//extern int gfx_display_columns;

extern int gfx_width;
extern int gfx_height;
extern int _gfx_width;
extern int _gfx_height;
extern int interlaceratio;
extern float fps_value;
//extern int vesa_scrolling_pan;
//extern int vesa_scrolling_origin_x;
//extern int vesa_scrolling_origin_y;

extern int horizontal_sync_delay;
extern int horizontal_disp_delay;

extern int pcmonitorlow_h;
extern int pcmonitorhigh_h;
extern int pcmonitorlow_v;
extern int pcmonitorhigh_v;


extern int usepalonly;
extern int usestrict15KHzonly;
extern int usespecial15KHz;
extern int use24KHz;
extern int use31KHz;
extern int usepcmonitor;

extern int use_svga15khz_panning;


extern SCRLIST ScrModeList[];
extern SCRLIST AutoResolutionList[];




static void SetMiscParam(CRTCPARAM *cp);
static void AdjustCenter(CRTCPARAM *cp);















int GetMode15KHz(CRTCPARAM *cp)
{
	SCRLIST *ListPtr;
	DWORD scrmode;
	int found;
	int idx;
	int display_columns;
	int display_lines;
	DWORD orientation;
	int vsyncflag;
#if 1
	int xoffset;
	int yoffset;
#endif
	int vesares_x;
	int vesares_y;
	long frame_rate;

	vesares_x = width_for_mode15k;
	vesares_y = height_for_mode15k;

	vsyncflag = video_sync | use_triplebuf;
	if(fps_value == 0.0){
		frame_rate = (long)((video_fps * 100) + 0.005);
		logerror("frame_rate = video_fps (= %.2f)\n",
									(float)frame_rate/100);
	}else{
		frame_rate = (long)((fps_value * 100) + 0.005);
		logerror("frame_rate = fps_value (= %.2f)\n",
									(float)frame_rate/100);
	}

	if((Machine->orientation & ORIENTATION_SWAP_XY) == 0)
		orientation = SCREEN_HORZ;
	else
		orientation = SCREEN_VERT;

	scrmode = 0;
	while(1){
		if(usepcmonitor != 0){
			scrmode = PCMONITOR;
			logerror("Using pc monitor\n");
			break;
		}
		if(usestrict15KHzonly != 0){
			scrmode = STRICT15KHz;
			logerror("Using strict 15KHz only\n");
			break;
		}
		if(usepalonly != 0){
			scrmode = PAL15KHz;
			logerror("Using PAL only\n");
			break;
		}

		logerror("Using ");
		if(use31KHz != 0){
			scrmode |= MODE31KHz;
			logerror("31KHz ");
		}
		if(use24KHz != 0){
			scrmode |= MODE24KHz | STRICT24KHz;
			logerror("24KHz ");
		}
		if(usespecial15KHz != 0){
			scrmode |= SPECIAL15KHz;
			logerror("SPECIAL15KHz ");
		}
		scrmode |= MODE15KHz | STRICT15KHz;
		logerror("15KHz ");
		logerror("\n");
		break;
	}

	logerror("_gfx_width.... %d\n", _gfx_width);
	logerror("_gfx_height... %d\n", _gfx_height);

	if(_gfx_width == 0 && _gfx_height == 0){
		display_columns = vesares_x;
		display_lines	= vesares_y;
	}else{
		display_columns = _gfx_width;
		display_lines	= _gfx_height;
	}


	if(vsyncflag != 0)
		logerror("vsync is on\n");

	ListPtr = ScrModeList;
	idx = 0;
	found = 0;
	if(found == 0){
		while(1){
			if( (ListPtr + idx)->forx == 0 || (ListPtr + idx)->fory == 0) {
				logerror("%dx%d @ %.2fHz mode not found\n",
											display_columns, display_lines,
											(float)frame_rate/100);
				if(vsyncflag != 0){
					logerror("vsync off\n");
					vsyncflag = 0;
					idx = 0;
					continue;
				}else{
					break;
				}
			}
			if( ((ListPtr + idx)->forx == display_columns  &&
				 (ListPtr + idx)->fory == display_lines    &&
				 (((ListPtr + idx)->group & scrmode) != 0) &&
				 (((ListPtr + idx)->orientation & orientation)) != 0)
			  )
			{
				if( vsyncflag == 0){
					logerror("Using %dx%d  H:%.2fKHz V:%.2fHz\n",
									(ListPtr + idx)->forx,
									(ListPtr + idx)->fory,
									(float)(ListPtr + idx)->cp->horzHz / 1000,
									(float)(ListPtr + idx)->cp->vert001Hz / 100);
					found = 1;
					break;
				}else{
					long ld0;

					ld0 = (ListPtr + idx)->cp->vert001Hz;
					if((ListPtr + idx)->cp->interlace != 0)
						ld0 /= 2;
					ld0 = abs(ld0 - frame_rate);
					if(ld0 < 10){
						logerror("Using %dx%d  H:%.2fKHz V:%.2fHz\n",
										(ListPtr + idx)->forx,
										(ListPtr + idx)->fory,
										(float)(ListPtr + idx)->cp->horzHz / 1000,
										(float)(ListPtr + idx)->cp->vert001Hz / 100);
						found = 1;
						break;
					}
				}
			}
			idx += 1;
		}
	}

	if(found == 0){
		logerror("AutoRes\n");
		ListPtr = AutoResolutionList;
		idx = 0;
		while(1){
			if( (ListPtr + idx)->forx == 0 || (ListPtr + idx)->fory == 0) {
				logerror("not found\n");
				return 0;
			}
			if( ((ListPtr + idx)->forx >= display_columns  &&
				 (ListPtr + idx)->fory >= display_lines    &&
				 (((ListPtr + idx)->group & scrmode) != 0) &&
				 (((ListPtr + idx)->orientation & orientation)) != 0)
			  )
			{
				logerror("Using %dx%d  H:%.2fKHz V:%.2fHz\n",
								(ListPtr + idx)->forx,
								(ListPtr + idx)->fory,
								(float)(ListPtr + idx)->cp->horzHz / 1000,
								(float)(ListPtr + idx)->cp->vert001Hz / 100);
				found = 1;
				break;
			}
			idx += 1;
		}
	}

#if 1
	xoffset = ((ListPtr + idx)->cp->width  - width_for_mode15k)  / 2;
	yoffset = ((ListPtr + idx)->cp->height - height_for_mode15k) / 2;
	if( (gfx_xoffset - xoffset) < 0 ){
		xoffset = 0;
	}else{
		xoffset = gfx_xoffset - xoffset;
	}
	if( (gfx_yoffset - yoffset) < 0 ){
		yoffset = 0;
	}else{
		yoffset = gfx_yoffset - yoffset;
	}
#endif

	logerror("orientation ");
	if(orientation == SCREEN_HORZ)
		logerror("Horz\n");
	else
		logerror("Vert\n");
	logerror("req_width........... %d\n", width_for_mode15k);
	logerror("req_height.......... %d\n", height_for_mode15k);
#if 1
	logerror("gfx_xoffset......... %d\n", gfx_xoffset);
	logerror("gfx_yoffset......... %d\n", gfx_yoffset);
	logerror("xoffset............. %d\n", xoffset);
	logerror("yoffset............. %d\n", yoffset);
#endif
	logerror("x................... %d\n", (ListPtr + idx)->cp->width);
	logerror("y................... %d\n", (ListPtr + idx)->cp->height);

	*cp = *((ListPtr + idx)->cp);
	AdjustCenter(cp);
	SetMiscParam(cp);
#if 1
	vesa_scroll(xoffset, yoffset);
//	  vesa_scrolling_origin_x = xoffset;
//	  vesa_scrolling_origin_y = yoffset;
#endif

	setinterlaceflag(0);

	while(usepcmonitor != 0){
		long vert001Hz;
		long horzHz;
		long dotclockHz;
		int limitlow_v;
		int limithigh_v;
		int limitlow_h;
		int limithigh_h;

		if((pcmonitorlow_v == 0) || (pcmonitorhigh_v == 0) ||
		   (pcmonitorlow_h == 0) || (pcmonitorhigh_h == 0))
		{
			limitlow_v	= 50;					// Hz
			limithigh_v = 85;					// Hz
			limitlow_h	= 24;					// KHz
			limithigh_h = 60;					// KHz
		}else{
			limitlow_v	= pcmonitorlow_v;		// Hz
			limithigh_v = pcmonitorhigh_v;		// Hz
			limitlow_h	= pcmonitorlow_h;		// KHz
			limithigh_h = pcmonitorhigh_h;		// KHz
		}

		logerror("PC MONITOR MODE\n");
		vert001Hz  = frame_rate;
		logerror("V %dHz\n", vert001Hz/100);
		logerror("pc monitor low  limit V %dHz\n", limitlow_v);
		logerror("pc monitor high limit V %dHz\n", limithigh_v);
		while((vert001Hz/100) <= limithigh_v){
			if( (limitlow_v <= (vert001Hz/100)) &&
				((vert001Hz/100) <= limithigh_v) )
			{
				break;
			}
			vert001Hz *= 2;
			logerror("x2 V rate\n");
		}
		if((vert001Hz/100) > limithigh_v){
			logerror("V %dHz > %dHz\n", vert001Hz/100, limithigh_v);
			logerror("Using default 60Hz mode\n");
			break;
		}
		dotclockHz = CalcDotclock(vert001Hz, cp->HTotal, cp->VTotal,
												cp->interlace);
		horzHz	   = CalcHorizontalHz(dotclockHz, cp->HTotal);
		logerror("H %dKHz\n", horzHz/1000);
		logerror("pc monitor low  limit H %dKHz\n", limitlow_h);
		logerror("pc monitor high limit H %dKHz\n", limithigh_h);
		if( !((limitlow_h <= (horzHz/1000)) && ((horzHz/1000) <= limithigh_h)) ){
			logerror("Using default 60Hz mode\n");
			break;
		}
		cp->vert001Hz  = vert001Hz;
		cp->horzHz	   = horzHz;
		cp->dotclockHz = dotclockHz;
		logerror("Set H:%.2fKHz V:%.2fHz dotclock %.4f MHz\n",
						(float)cp->horzHz / 1000,
						(float)cp->vert001Hz / 100,
						(float)(cp->dotclockHz / 1000) / 1000.0);
		break;
	}


	gfx_width  = cp->width;
	gfx_height = cp->height;

	use_svga15khz_panning = 1;

	return 1;
}




static void vesa_scroll(int x, int y)
{
	__dpmi_regs _dpmi_reg;

	_dpmi_reg.x.ax = 0x4F07;
	_dpmi_reg.x.bx = 0;
	_dpmi_reg.x.cx = x;
	_dpmi_reg.x.dx = y;
	__dpmi_int(0x10, &_dpmi_reg);
}



static void SetMiscParam(CRTCPARAM *cp)
{
//	  vesa_scrolling_pan = 1;

	if(interlaceratio != 0){
		if(interlaceratio < 0)	 interlaceratio = 0;
		if(interlaceratio > 200) interlaceratio = 200;
		cp->interlaceratio = interlaceratio;
	}
}

static void AdjustCenter(CRTCPARAM *cp)
{
	int d0;

	if(center_x < 0){
		if((cp->HSStart + center_x) < (cp->HDisp + 8)){
			d0 = cp->HSStart;
			cp->HSStart = cp->HDisp + 8;
			cp->HSEnd  -= (d0 - cp->HSStart);
		}else{
			cp->HSStart = cp->HSStart + center_x;
			cp->HSEnd	= cp->HSEnd   + center_x;
		}
	}else{
		if((cp->HSEnd + center_x) > (cp->HTotal - 16)){
			d0 = cp->HSEnd;
			cp->HSEnd	 = cp->HTotal - 16;
			cp->HSStart += (cp->HSEnd - d0);
		}else{
			cp->HSEnd	= cp->HSEnd   + center_x;
			cp->HSStart = cp->HSStart + center_x;
		}
	}

	if(center_y < 0){
		if((cp->VSStart + center_y) < (cp->VDisp + 1)){
			d0 = cp->VSStart;
			cp->VSStart = cp->VDisp + 1;
			cp->VSEnd  -= (d0 - cp->VSStart);
		}else{
			cp->VSStart = cp->VSStart + center_y;
			cp->VSEnd	= cp->VSEnd   + center_y;
		}
	}else{
		if((cp->VSEnd + center_y) > (cp->VTotal - 2)){
			d0 = cp->VSEnd;
			cp->VSEnd	 = cp->VTotal - 2;
			cp->VSStart += (cp->VSEnd - d0);
		}else{
			cp->VSEnd	= cp->VSEnd   + center_y;
			cp->VSStart = cp->VSStart + center_y;
		}
	}
}


void Centering(CRTCPARAM *cp, int dx, int dy)
{
	if(dx < 0){
		if((cp->HSStart + dx) > cp->HDisp){
			cp->HSStart += dx;
			cp->HSEnd	+= dx;
		}
	}else{
		if((cp->HSEnd + dx) < cp->HTotal){
			cp->HSStart += dx;
			cp->HSEnd	+= dx;
		}
	}

	if(dy < 0){
		if((cp->VSStart + dy) > cp->VDisp){
			cp->VSStart += dy;
			cp->VSEnd	+= dy;
		}
	}else{
		if((cp->VSEnd + dy) < cp->VTotal){
			cp->VSStart += dy;
			cp->VSEnd	+= dy;
		}
	}
}




#if 0
void GetCrtcParam(CRTCPARAM *cp, MiscParam *mc, _Register *rg, int registerlen)
{
	BYTE port[0x20];
	int d0;
	int d1;

	d0 = 0;
	while(d0 < registerlen){
		port[ rg[d0].index ] = rg[d0].value;
		d0 += 1;
	}


/*										*/
/*				 CRTC					*/
/*										*/

	cp->HTotal	= (port[0x00] + 5) * 8;
	cp->HDisp	= (port[0x01] + 1) * 8;

	d0 = port[0x02];
	cp->HBStart = d0 * 8;
	d1 = (port[0x03] & 0x1F) | BitMov(0, 5, port[0x05], 7);
	while( (d0 & 0x3F) != (d1 & 0x3F) ){
		d0 += 1;
	}
	cp->HBEnd	= d0 * 8;

	d0 = port[0x04];
//	  cp->HSStart = d0 * 8;
	cp->HSStart = d0 * 8 + mc->hskew;
	d1 = (port[0x05] & 0x1F);
	while( (d0 & 0x1F) != (d1 & 0x1F) ){
		d0 += 1;
	}
	cp->HSEnd	= d0 * 8;

	cp->VTotal	= (port[0x06] | BitMov(0, 8, port[0x07], 0) |
								BitMov(0, 9, port[0x07], 5) ) + 2;
	cp->VDisp	= (port[0x12] | BitMov(0, 8, port[0x07], 1) |
								BitMov(0, 9, port[0x07], 6) ) + 1;
	cp->VBStart = (port[0x15] | BitMov(0, 8, port[0x07], 3) |
								BitMov(0, 9, port[0x09], 5) );
	d0 = cp->VBStart;
	d1 = port[0x16];
	while( (d0 & 0xFF) != (d1 & 0xFF) ){
		d0 += 1;
	}

//.    cp->VBEnd = d0;
//.    cp->VBEnd = cp->VTotal - 16;


	cp->VSStart = (port[0x10] | BitMov(0, 8, port[0x07], 2) |
								BitMov(0, 9, port[0x07], 7) );
	d0 = cp->VSStart;
	d1 = port[0x11] & 0x0F;
	while( (d0 & 0x0F) != (d1 & 0x0F) ){
		d0 += 1;
	}
	cp->VSEnd = d0;
//.
	cp->VBEnd = cp->VSEnd + 1;
//.



/*										*/
/*				 MISC					*/
/*										*/

	cp->width  = mc->x;
	cp->height = mc->y;
	cp->depth  = 0;

	cp->dotclockHz = (long)(mc->dotclock * 1000 * 1000);
	cp->horzHz	   = (long)(mc->horzKHz * 1000);
	cp->vert001Hz  = (long)((mc->vertHz + 0.005) * 100);

	cp->hpolarity  = mc->hpolarity;
	cp->vpolarity  = mc->vpolarity;
	cp->doublescan = (port[0x09] >> 7) & 0x01;
	cp->interlace  = mc->interlace;

	if(interlaceratio < 0)	 interlaceratio = 0;
	if(interlaceratio > 200) interlaceratio = 200;
	cp->interlaceratio = interlaceratio;
	cp->compositesync  = compositesync;

	cp->sis_numerator	= mc->numerator;
	cp->sis_denumerator = mc->denumerator;
	cp->sis_divider 	= mc->divider;
	cp->sis_postscalar	= mc->postscalar;
	cp->sis_seqdivider	= mc->seqdivider;
}
#endif




void SetGenericVGARegister(CRTCPARAM *cp)
{
	int d0;
	int d1;
	int HDispSkew;
	int HSyncDelay;

	HSyncDelay = horizontal_sync_delay & 0x03;
	HDispSkew  = horizontal_disp_delay & 0x03;

	// Unlock
	SetCrtcReg(0x11, BitMov(GetCrtcReg(0x11), 7, 0, 0));

	// HTotal
	d0 = cp->HTotal;
	d0 = (d0 / 8) - 5;
	SetCrtcReg(0x00, d0 & 0xFF);

	// HDisp
	d0 = cp->HDisp;
	d0 = (d0 / 8) - 1;
	SetCrtcReg(0x01, d0 & 0xFF);

	// HBlankStart
	d0 = cp->HBStart;
	d0 = d0 / 8;
	SetCrtcReg(0x02, d0 & 0xFF);

	// HBlankEnd
	d0 = cp->HBEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x03);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HDispSkew, 0);	// horizontal disp skew
	d1 = BitMov(d1, 6, HDispSkew, 1);	// horizontal disp skew
	d1 = BitMov(d1, 7, 1, 0);			// compatible read
	SetCrtcReg(0x03, d1);
	SetCrtcReg(0x05, BitMov(GetCrtcReg(0x05), 7, d0, 5));

	// HSyncStart
	d0 = cp->HSStart;
	d0 = d0 / 8;
	SetCrtcReg(0x04, d0 & 0xFF);

	// HSyncEnd
	d0 = cp->HSEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x05);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HSyncDelay, 0);
	d1 = BitMov(d1, 6, HSyncDelay, 1);
	SetCrtcReg(0x05, d1);

	// VTotal
	d0 = cp->VTotal;
	d0 = d0 - 2;
	SetCrtcReg(0x06, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 0, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 5, d0, 9));

	// VDisp
	d0 = cp->VDisp;
	d0 = d0 - 1;
	SetCrtcReg(0x12, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 1, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 6, d0, 9));


	// VBlankStart
	d0 = cp->VBStart;
	SetCrtcReg(0x15, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 3, d0, 8));
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 5, d0, 9));

	// VBlankEnd
	d0 = cp->VBEnd;
	SetCrtcReg(0x16, d0 & 0xFF);

	// VSyncStart
	d0 = cp->VSStart;
	SetCrtcReg(0x10, d0 & 0xFF);
	d1 = GetCrtcReg(0x07);
	d1 = BitMov(d1, 2, d0,	8);  // bit 8
	d1 = BitMov(d1, 7, d0,	9);  // bit 9
	SetCrtcReg(0x07, d1);

	// VSyncEnd
	d0 = cp->VSEnd;
	d0 = d0 & 0x0F;
	d1 = GetCrtcReg(0x11) & 0xF0;
	d1 = d1 | d0;
	SetCrtcReg(0x11, d1);

	// Double Scan Dis.
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));

}



long CalcDotclock(long Vert001Hz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = f0 * Vert001Hz;
	f0 = f0 / 100.0;

	return (long)f0;
}

long CalcHorizontalHz(long dotclockHz, int HTotal)
{
	return dotclockHz / HTotal;
}

long CalcVert001Hz(long dotclockHz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = dotclockHz / f0;
	f0 = f0 * 100.0;

	return (long)f0;
}



void DispCrtcParam(CRTCPARAM *cp)
{
	logerror("DotClock..... %.4lf MHz\n",
						(double)(cp->dotclockHz / 1000.0) / 1000.0);
	logerror("HTotal....... %d\n", cp->HTotal);
	logerror("HDispEnd..... %d\n", cp->HDisp);
	logerror("HBlankStart.. %d\n", cp->HBStart);
	logerror("HSyncStart... %d\n", cp->HSStart);
	logerror("HSyncEnd..... %d\n", cp->HSEnd);
	logerror("HBlankEnd.... %d\n", cp->HBEnd);
	logerror("VTotal....... %d\n", cp->VTotal);
	logerror("VDispEnd..... %d\n", cp->VDisp);
	logerror("VBlankStart.. %d\n", cp->VBStart);
	logerror("VSyncStart... %d\n", cp->VSStart);
	logerror("VSyncEnd..... %d\n", cp->VSEnd);
	logerror("VBlankEnd.... %d\n", cp->VBEnd);
	logerror("Horz......... %.2lf KHz\n",
						(double)cp->horzHz / 1000.0);
	logerror("Vert......... %.2lf Hz\n",
						(double)cp->vert001Hz / 100.0);

	if(cp->hpolarity == 0){
		logerror("HorzSync..... Positive (Low  -> High)\n");
	}else{
		logerror("HorzSync..... Negative (High -> Low)\n");
	}
	if(cp->vpolarity == 0){
		logerror("VertSync..... Positive (Low  -> High)\n");
	}else{
		logerror("VertSync..... Negative (High -> Low)\n");
	}

	if(cp->interlace == 0){
		logerror("non interlaced mode\n");
	}else{
		logerror("interlaced mode\n");
	}
	if(cp->doublescan != 0){
		logerror("double scan\n");
	}

}


void SetHVPolarity(int HPolar, int VPolar)
{
	BYTE d0;

	if(HPolar != 0) HPolar = 1;
	if(VPolar != 0) VPolar = 1;

	d0 = _in(0x3CC);
	d0 = d0 & 0x3F;
	d0 = d0 | (VPolar << 7) | (HPolar << 6);
	_out(0x3C2, d0);
}





DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos)
{
	DWORD d0;
	DWORD d1;

	d1 = srcdata >> srcbitpos;
	d1 = d1 & 1;
	d0 = dstdata & ~(1 << dstbitpos);
	d1 = d1 << dstbitpos;
	d0 = d0 | d1;
	return d0;
}



void SetSeqReg(BYTE __idx, BYTE data)
{
	_out(0x3c4, __idx);
	_out(0x3c5, data);
}

BYTE GetSeqReg(BYTE __idx)
{
	BYTE d0;

	_out(0x3c4, __idx);
	d0 = _in(0x3c5);
	return d0;
}


BYTE GetAtrbData(BYTE __idx)
{
	_in(0x3da);
	_out(0x3c0, __idx | (1 << 5));
	return _in(0x3c1);
}

void SetAtrbData(BYTE __idx, BYTE data)
{
	_in(0x3da);
	_out(0x3c0, __idx | 0x20);
	_out(0x3c0, data);
}

int GetCrtcAddress(void)
{
	if((_inb(0x3CC) & 0x01) == 0)
		return 0x3B4;
	else
		return 0x3D4;
}

void SetCrtcReg(BYTE __idx, BYTE data)
{
	_outb(GetCrtcAddress(),    __idx);
	_outb(GetCrtcAddress() + 1, data);
}

BYTE GetCrtcReg(BYTE __idx)
{
	BYTE d0;

	_outb(GetCrtcAddress(), __idx);
	d0 = _inb(GetCrtcAddress() + 1);
	return d0;
}



int GetSeqDivider(void)
{
	int d0;

	d0	= (GetSeqReg(0x01) & 0x08) >>3;
	d0	= d0 + 1;
	return d0;
}

void SetSeqDivider(int d0)
{
	int d1;

	d0 -= 1;
	if(d0 != 0) d0 = 1 << 3;
	d1 = GetSeqReg(0x01) & ~0x08;
	d1 = d1 | d0;
	SetSeqReg(0x01, d1);
}


void _out(int edx, BYTE al)
{
	outportb(edx, al);
}
BYTE _in(int edx)
{
	return inportb(edx);
}

void _outb(int edx, BYTE al)
{
	outportb(edx, al);
}
BYTE _inb(int edx)
{
	return inportb(edx);
}

void _outw(int edx, WORD ax)
{
	outportw(edx, ax);
}
WORD _inw(int edx)
{
	return inportw(edx);
}

void _outl(int edx, DWORD eax)
{
	outportl(edx, eax);
}
DWORD _inl(int edx)
{
	return inportl(edx);
}


void DisableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 & 0x7F;
	SetCrtcReg(0x17, d0);
}

void EnableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 | 0x80;
	SetCrtcReg(0x17, d0);
}

void DisableStretch(void)
{
	stretch = 0;
	scanlines = 0;
}


static CRTCPARAM rcp;

void RegCRTCPARAM(CRTCPARAM *cp)
{
	rcp = *cp;
}

void GetCRTCPARAM(CRTCPARAM *cp)
{
	*cp = rcp;
}
