#include "mamalleg.h"
#include "driver.h"
#include "ticker.h"
#include <time.h>


//.s
void CLIB_DECL logerror(const char *text,...);
//.e


#define USE_UCLOCK
//#define USE_RDTSC

static int use_rdtsc;
TICKER ticks_per_sec;

TICKER ticker(void)
{
	if (use_rdtsc)
	{
		INT64 result;

		__asm__ __volatile__ (
			"rdtsc"
			: "=A" (result)
		);

		return result;
	}
	else
	{
		/* this assumes that uclock_t is 64-bit (which it is) */
		return uclock();
	}
}

int cpu_rdtsc(void)
{
	int result;

	__asm__ (
		"movl $1,%%eax	   ; "
		"xorl %%ebx,%%ebx  ; "
		"xorl %%ecx,%%ecx  ; "
		"xorl %%edx,%%edx  ; "
		"cpuid			   ; "
		"testl $0x10,%%edx ; "
		"setne %%al 	   ; "
		"andl $1,%%eax	   ; "
	:  "=&a" (result)	/* the result has to go in eax */
	:		/* no inputs */
	:  "%ebx", "%ecx", "%edx" /* clobbers ebx ecx edx */
	);
	return result;
}

void init_ticker(void)
{
	/* if the RDTSC instruction is available use it because */
	/* it is more precise and has less overhead than uclock() */
	if (cpu_cpuid && cpu_rdtsc())
	{
		uclock_t a,b;
		TICKER start,end;
		TICKER result[2];
		TICKER temp;
		int i;
		int limit;

		use_rdtsc = 1;	/* must set this before calling ticker() */
		limit = 16;
		while(limit > 0){
			i = 0;
			while(i < 2){
				a = uclock();
				/* wait some time to let everything stabilize */
				do
				{
					b = uclock();
				} while (b-a < UCLOCKS_PER_SEC/10);

				a = uclock();
				start = ticker();
				do
				{
					b = uclock();
				} while (b-a < UCLOCKS_PER_SEC/4);
				end = ticker();
				result[i] = (end - start)*4;
				i += 1;
			}
			if(result[0] > result[1]){
				temp = result[0] - result[1];
			}else{
				temp = result[1] - result[0];
			}
			logerror("result[0]  %d.%06d MHz\n",
					(int)(result[0]/1000000),(int)(result[0]%1000000));
			logerror("result[1]  %d.%06d MHz\n",
					(int)(result[1]/1000000),(int)(result[1]%1000000));
			if( temp < 1000000/2 ) break;
			logerror("retry\n");
			limit -= 1;
		}
		ticks_per_sec = (result[0] + result[1]) / 2;

		logerror("using RDTSC for timing, CPU speed = %d.%06d MHz\n",
				(int)(ticks_per_sec/1000000),(int)(ticks_per_sec%1000000));
	}
	else
	{
		use_rdtsc = 0;
		ticks_per_sec = UCLOCKS_PER_SEC;

		logerror("using uclock() for timing\n");
	}
}
